from typing import List, Union

from imap_tools import BaseMailBox, MailMessage
from pydantic import Field

from .action import Action, MoveAction
from .common import BaseModel
from .condition import Condition


class Rule(BaseModel):
    condition: Condition
    move_to: Union[str, None]
    keep_evaluating: bool = False
    actions: List[Union[str, Action]] = Field(default_factory=list)

    def apply(self, mailbox: BaseMailBox, message: MailMessage, dry_run: bool):
        if self.condition.eval(message):
            for action in self._actions():
                action.apply(mailbox, message, dry_run)
            return True
        return False

    def _actions(self) -> List[Action]:
        from ..config_state import state

        res = []
        for action in self.actions:
            if isinstance(action, str):
                res.append(state.actions[action])
            else:
                res.append(action)

        if self.move_to is not None:
            res.append(MoveAction(dest=self.move_to))
        return res

    def destinations(self):
        return [action.dest for action in self._actions() if action.action == "move"]


def apply_list(rules: List[Rule], mailbox: BaseMailBox, message: MailMessage, dry_run: bool):
    for rule in rules:
        applied = rule.apply(mailbox, message, dry_run)
        if applied and not rule.keep_evaluating:
            break



import typer
from rich import print
from rich.pretty import Node, pretty_repr

from ..config_state import state


def setup_callback(config_file: typer.FileText = typer.Option(...)):
    state.parse_config(config_file)
    state.create_mailbox()


app = typer.Typer(
    name="folder",
    callback=setup_callback,
    no_args_is_help=True,
    help="Various IMAP folder helpers.",
)


def render_with_name(name, value):
    value_repr = pretty_repr(value)
    return Node(key_repr=name, value_repr=value_repr).render()


def print_with_name(name, value):
    print(render_with_name(name, value))


@app.command(rich_help_panel="Generic commands")
def list():
    """List existing folders."""
    folder_names = sorted((f.name for f in state.mailbox.folder.list()))
    print_with_name("folder_names", folder_names)


@app.command(rich_help_panel="Generic commands")
def create(name: str):
    """Create a new folder."""
    state.mailbox.folder.create(name)


@app.command(rich_help_panel="Generic commands")
def delete(name: str):
    """Delete a folder."""
    state.mailbox.folder.delete(name)


@app.command(rich_help_panel="Generic commands")
def rename(old_name: str, new_name: str):
    """Rename a folder."""
    state.mailbox.folder.rename(old_name, new_name)


@app.command(rich_help_panel="Helper commands")
def show_destinations():
    """List destinations folders from your configuration."""
    destinations = sorted({dest for rule in state.rules for dest in rule.destinations()})
    print_with_name("[bold]destinations", destinations)


@app.command(rich_help_panel="Helper commands")
def compare_destinations():
    """Compare existing folders with destinations from the configuration"""
    destinations = {dest for rule in state.rules for dest in rule.destinations()}
    folder_names = {f.name for f in state.mailbox.folder.list()}
    folders_with_destination = sorted(destinations & folder_names)
    destinations_without_folder = sorted(destinations - folder_names)
    folders_without_destination = sorted(folder_names - destinations)
    print_with_name("[bold]folders_with_destination", folders_with_destination)
    print_with_name("[bold]folders_without_destination", folders_without_destination)
    print_with_name("[bold]destinations_without_folder", destinations_without_folder)


@app.command(rich_help_panel="Helper commands")
def create_missing_folders():
    """Create missing folders."""
    destinations = {dest for rule in state.rules for dest in rule.destinations()}
    folder_names = {f.name for f in state.mailbox.folder.list()}
    destinations_without_folder = sorted(destinations - folder_names)

    print_with_name("will_create_folders", destinations_without_folder)

    input("Will create the given folders, is that oK? (Ctrl-c to quit)")
    for folder in destinations_without_folder:
        state.mailbox.folder.create(folder)



import logging
import os
import ssl
import sys
from typing import List

import certifi
import yaml
from imap_tools import BaseMailBox, MailBox

from mmuxer.models.action import Action, ActionLoader, DeleteAction, FlagAction, MoveAction
from mmuxer.models.enums import Flag
from mmuxer.models.rule import Rule
from mmuxer.models.settings import Settings

logger = logging.getLogger(__name__)


def make_ssl_context(ssl_ciphers):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    if ssl_ciphers is not None:
        context.set_ciphers(ssl_ciphers)
    context.load_verify_locations(
        cafile=os.path.relpath(certifi.where()), capath=None, cadata=None
    )
    return context


default_actions = {
    "delete": DeleteAction(),
    "trash": MoveAction(dest="Trash"),
    "mark_read": FlagAction(flag=Flag.SEEN),
}


class State:
    def __init__(self):
        self._settings = None
        self._rules = None
        self._mailbox = None
        self.actions: dict[str, Action] = default_actions

    def create_mailbox(self):
        logger.debug(f"Connecting to {self.settings.server} with {self.settings.username}")
        ssl_context = make_ssl_context(self.settings.ssl_ciphers)
        self._mailbox = MailBox(self.settings.server, ssl_context=ssl_context).login(
            self.settings.username, self.settings.password
        )
        logger.info(f"Connected to {self.settings.server} with {self.settings.username}")

    def parse_config(self, config_raw):
        try:
            config_dict = yaml.safe_load(config_raw)
        except Exception:
            logger.exception("Failed loading config file, check that the yaml syntax is valid.")
            raise

        if (
            not isinstance(config_dict, dict)
            or (config_dict.keys() - {"rules", "settings", "actions"}) != set()
            or "rules" not in config_dict
            or "settings" not in config_dict
        ):
            logger.error(
                "The config file should contain a mapping with the keys 'rules', 'settings' and 'actions' (optional)."
            )
            sys.exit(1)

        try:
            self._settings = Settings(**config_dict["settings"])
        except Exception as exc:
            logger.error("Failed parsing the 'settings' section of the configuration:")
            logger.error(exc)
            sys.exit(1)

        parsed_rules = []
        for rule_dict in config_dict["rules"]:
            try:
                parsed_rules.append(Rule.parse_obj(rule_dict))
            except Exception as exc:
                logger.error("Failed parsing the 'rules' section of the configuration:")
                logger.error(rule_dict)
                logger.error(exc)
                sys.exit(1)
        self._rules = parsed_rules

        if "actions" in config_dict:
            if not isinstance(config_dict["actions"], dict):
                logger.error(f"'actions' value is not a mapping: {config_dict['actions']}")
                sys.exit(1)
            for action_name, action_dict in config_dict["actions"].items():
                if action_name in self.actions:
                    logger.warning(f"Overriding default action f{action_name}")
                try:
                    self.actions[action_name] = ActionLoader.parse_obj(action_dict).__root__
                except Exception as exc:
                    logger.error("Failed parsing the 'actions' section of the configuration:")
                    logger.error(action_dict)
                    logger.error(exc)
                    sys.exit(1)
            self._rules = parsed_rules

    @property
    def settings(self) -> Settings:
        if self._settings is None:
            raise Exception("Uninitialized settings")
        return self._settings

    @property
    def rules(self) -> List[Rule]:
        if self._rules is None:
            raise Exception("Uninitialized rules")
        return self._rules

    @property
    def mailbox(self) -> BaseMailBox:
        if self._mailbox is None:
            raise Exception("Uninitialized mailbox")
        return self._mailbox


state = State()



import logging
from abc import abstractmethod
from typing import Literal, Union

from imap_tools import BaseMailBox, MailMessage

from .common import BaseModel
from .enums import Flag

logger = logging.getLogger(__name__)


def format_message(msg: MailMessage):
    return f"[{{{msg.uid}}} {msg.from_} -> {msg.to} '{msg.subject}']"


class BaseAction(BaseModel):
    @abstractmethod
    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        pass

    @abstractmethod
    def format(self, message):
        pass

    def apply(self, mailbox: BaseMailBox, message: MailMessage, dry_run: bool):
        logger.info(self.format(message))
        if not dry_run:
            logger.info(self.format(message))
            self._apply(mailbox, message)


class MoveAction(BaseAction):
    action: Literal["move"] = "move"
    dest: str

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.move(message.uid, self.dest)

    def format(self, message):
        return f"MOVE {format_message(message)} --> {self.dest}"


class DeleteAction(BaseAction):
    action: Literal["delete"] = "delete"

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.delete([message.uid])

    def format(self, message):
        return f"DELETE {format_message(message)}"


class FlagAction(BaseAction):
    action: Literal["flag"] = "flag"
    flag: Flag

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.flag([message.uid], flag_set={self.flag.imap}, value=True)

    def format(self, message):
        return f"FLAG {format_message(message)} {self.flag.name}"


class UnflagAction(BaseAction):
    action: Literal["flag"] = "flag"
    flag: Flag

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.flag([message.uid], flag_set={self.flag.imap}, value=False)

    def format(self, message):
        return f"UNFLAG {format_message(message)} {self.flag.name}"


Action = Union[MoveAction, DeleteAction, FlagAction, UnflagAction]


class ActionLoader(BaseModel):
    __root__: Action



from enum import Enum


class ComparisonOperator(Enum):
    CONTAINS = "CONTAINS"
    EQUALS = "EQUALS"


class Flag(Enum):
    SEEN = "SEEN"
    ANSWERED = "ANSWERED"
    FLAGGED = "FLAGGED"
    DELETED = "DELETED"
    DRAFT = "DRAFT"

    @property
    def imap(self):
        return "\\" + self.name.capitalize()



import logging
from abc import abstractmethod
from typing import ForwardRef, List, Union

from imap_tools import MailMessage

from .common import BaseModel
from .enums import ComparisonOperator


class IBaseCondition(BaseModel):
    operator: ComparisonOperator = ComparisonOperator.CONTAINS

    @abstractmethod
    def get_value(self, message: MailMessage) -> str:
        pass

    @abstractmethod
    def get_operand(self) -> str:
        pass

    def eval(self, message: MailMessage) -> bool:
        value = self.get_value(message)
        logging.debug("Eval %s <%s> %s", self.get_operand(), self.operator, value)
        if self.operator == ComparisonOperator.CONTAINS:
            return self.get_operand().lower() in value.lower()
        if self.operator == ComparisonOperator.EQUALS:
            return self.get_operand().lower() == value.lower()
        raise Exception(f"Unhandled operator {self.operator}")


class From(IBaseCondition):
    FROM: str

    def get_value(self, message: MailMessage):
        return message.from_

    def get_operand(self) -> str:
        return self.FROM

    def __rich_repr__(self):
        yield self.operator.name, self.FROM


class To(IBaseCondition):
    TO: str

    def get_value(self, message: MailMessage):
        return " ".join(message.to)

    def get_operand(self) -> str:
        return self.TO

    def __rich_repr__(self):
        yield self.operator.name, self.TO


class Subject(IBaseCondition):
    SUBJECT: str

    def get_value(self, message: MailMessage):
        return message.subject

    def get_operand(self) -> str:
        return self.SUBJECT

    def __rich_repr__(self):
        yield self.operator.name, self.SUBJECT


class Body(IBaseCondition):
    BODY: str

    def get_value(self, message: MailMessage):
        return message.text + message.html

    def get_operand(self) -> str:
        return self.BODY

    def __rich_repr__(self):
        yield self.operator.name, self.BODY


BaseCondition = Union[From, To, Subject]

All = ForwardRef("All")  # type: ignore
Any = ForwardRef("Any")  # type: ignore
Not = ForwardRef("Not")  # type: ignore


class All(BaseModel):
    ALL: List[Union[BaseCondition, All, Any, Not]]

    def eval(self, message: MailMessage):
        return all(operand.eval(message) for operand in self.ALL)

    def __rich_repr__(self):
        for item in self.ALL:
            yield item


class Any(BaseModel):
    ANY: List[Union[BaseCondition, All, Any, Not]]

    def eval(self, message: MailMessage):
        return any(operand.eval(message) for operand in self.ANY)

    def __rich_repr__(self):
        for item in self.ANY:
            yield item


class Not(BaseModel):
    NOT: Union[BaseCondition, All, Any, Not]

    def eval(self, message: MailMessage):
        return not self.NOT.eval(message)


All.update_forward_refs()
Any.update_forward_refs()
Not.update_forward_refs()

Condition = Union[BaseCondition, All, Any, Not]



from typing import Optional

from pydantic import BaseSettings


class Settings(BaseSettings):
    server: str
    username: str
    password: str
    ssl_ciphers: Optional[str] = None

    class Config:
        env_file = ".env"
        secrets_dir = "/run/secrets"



import imaplib
import logging

import typer
from imap_tools import AND

from ..config_state import state
from ..models.rule import apply_list

logger = logging.getLogger(__name__)


def tidy(
    folder: str = typer.Option(None, help="Folder to fetch the messages from"),
    dry_run: bool = typer.Option(False, help="Print actions instead of running them"),
    config_file: typer.FileText = typer.Option(...),
):
    """Run once, on all messages of the INBOX (or the given folder)."""
    state.parse_config(config_file)
    state.create_mailbox()
    box = state.mailbox
    if folder is not None:
        box.folder.set(folder)
    counter = 0
    for msg in box.fetch(bulk=True):
        apply_list(state.rules, box, msg, dry_run)
        counter += 1
    print()
    print(f"{counter} messages parsed.")


def monitor(
    folder: str = typer.Option(None, help="Folder to fetch the messages from"),
    dry_run: bool = typer.Option(False, help="Print actions instead of running them"),
    config_file: typer.FileText = typer.Option(...),
):
    """Monitor mailbox, and apply rules on unseen messages."""
    state.parse_config(config_file)
    state.create_mailbox()
    box = state.mailbox

    if folder is not None:
        box.folder.set(folder)
        logging.info(f"Starting monitoring {folder}")
    else:
        logging.info("Starting monitoring INBOX")
    while True:
        try:
            responses = box.idle.wait(timeout=60)
            if responses:
                for msg in box.fetch(AND(seen=False), mark_seen=False):
                    print(f"Found message [{{{msg.uid}}} {msg.from_} -> {msg.to} '{msg.subject}']")
                    apply_list(state.rules, box, msg, dry_run)
        except imaplib.IMAP4.abort:
            logger.warning("IMAP connection aborted, reconnecting ")
            state.create_mailbox()
            box = state.mailbox
        except Exception:
            logger.exception("An error occured, reconnecting...")
            state.create_mailbox()
            box = state.mailbox
